/////////////////////////////////////////////////////////////////////////////
// Name:        wxcard.cpp
// Purpose:     wxCard is a vCard editor/viewer written in wxWindows
// Author:      Vadim Zeitlin
// Modified by:
// Created:     31.07.00
// RCS-ID:      $Id: wxcard.cpp,v 1.16 2000/08/14 16:15:22 zeitlin Exp $
// Copyright:   (c) 2000 Vadim Zeitlin <vadim@wxwindows.org>
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
    #pragma implementation "wxcard.cpp"
    #pragma interface "wxcard.cpp"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

// for all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWindows headers)
#ifndef WX_PRECOMP
    #include <wx/app.h>
    #include <wx/intl.h>
    #include <wx/log.h>
    #include <wx/dialog.h>
    #include <wx/menu.h>
    #include <wx/toolbar.h>
    #include <wx/filedlg.h>
    #include <wx/msgdlg.h>
    #include <wx/file.h>
#endif

#include <wx/cmdline.h>
#include <wx/listctrl.h>
#include <wx/ffile.h>

#include "vcard.h"

// ----------------------------------------------------------------------------
// resources
// ----------------------------------------------------------------------------

// under Windows we use the resources
#ifndef __WXMSW__
    // the application icon
    #include "icons/wxcard.xpm"

    // the toolbar icons
    #include "icons/tb_open.xpm"
    #include "icons/tb_new.xpm"
    #include "icons/tb_delete.xpm"
    #include "icons/tb_about.xpm"
#endif // !Windows

// ----------------------------------------------------------------------------
// constants
// ----------------------------------------------------------------------------

// the parameters for file save/load dialogs
static const wxChar *VCARD_EXT = _T(".vcf");
static const wxChar *VCARD_NAME = _T("vCard");

// ----------------------------------------------------------------------------
// private classes
// ----------------------------------------------------------------------------

// Define a new application type, each program should derive a class from wxApp
class vCardApp : public wxApp
{
public:
    vCardApp() { m_locale = NULL; }

    // override base class virtuals
    // ----------------------------

    // this one is called on application startup and is a good place for the
    // app initialization (doing it here and not in the ctor allows to have an
    // error return: if OnInit() returns false, the application terminates)
    virtual bool OnInit();

    // the match to OnInit(): called before the application termination,
    // returns the program exit code
    virtual int OnExit();

private:
    // the locale object we use, NULL if the default (us_EN) is used
    wxLocale *m_locale;
};

// Define a new frame type: this is going to be our main frame
class vCardFrame : public wxFrame
{
public:
    // ctor and dtor
    vCardFrame();
    virtual ~vCardFrame();

    // load all vCards from the specified file
    void OpenFile(const wxString& filename);

    // save the current file if needed, return TRUE on success
    bool Save();

    // event handlers (these functions should _not_ be virtual)
    void OnNewFile(wxCommandEvent& event);
    void OnOpenFile(wxCommandEvent& event);
    void OnSave(wxCommandEvent& event);
    void OnQuit(wxCommandEvent& event);

    void OnAddCard(wxCommandEvent& event);
    void OnEditCard(wxCommandEvent& event);
    void OnDeleteCard(wxCommandEvent& event);

    void OnAbout(wxCommandEvent& event);

    void OnUpdateEditMenu(wxUpdateUIEvent& event);

    void OnListItemActivated(wxListEvent& event);

    void OnClose(wxCloseEvent& event);

protected:
    // set the current file name, update the title and the dirty flag
    void SetFilename(const wxString& filename);

    // return the current list ctrl selection
    long GetSelection() const;

    // save the current file unconditionally
    bool DoSave();

    // edit the card with the given index
    void DoEdit(long n);

    // delete the card with the given index
    void DoDelete(long n);

    // add a card to the list control
    bool DoAddCard(const wxVCard *vcard);

    // the array of cards in the current file
    wxArrayCards m_cards;

    // the list control holding vCards
    wxListCtrl *m_listctrl;

    // the name of the currently opened file
    wxString m_filename;

    // dirty flag: do we have anything to save?
    bool m_dirty;

private:
    // any class wishing to process wxWindows events must use this macro
    DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// constants
// ----------------------------------------------------------------------------

// IDs for the controls and the menu commands
enum
{
    // menu items
    vCard_File_New = 100,
    vCard_File_Open,
    vCard_File_Save,
    vCard_File_Quit,

    vCard_Edit_Add = 200,
    vCard_Edit_Edit,
    vCard_Edit_Delete,

    vCard_Help_About = 800,

    // controls
    vCard_ListCtrl = 1000
};

// the columns of the list ctrl
enum
{
    Column_Name,
    Column_Email,
    Column_Max
};

// ----------------------------------------------------------------------------
// event tables and other macros for wxWindows
// ----------------------------------------------------------------------------

// the event tables connect the wxWindows events with the functions (event
// handlers) which process them. It can be also done at run-time, but for the
// simple menu events like this the static method is much simpler.
BEGIN_EVENT_TABLE(vCardFrame, wxFrame)
    EVT_MENU(vCard_File_New,    vCardFrame::OnNewFile)
    EVT_MENU(vCard_File_Open,   vCardFrame::OnOpenFile)
    EVT_MENU(vCard_File_Save,   vCardFrame::OnSave)
    EVT_MENU(vCard_File_Quit,   vCardFrame::OnQuit)

    EVT_MENU(vCard_Edit_Add,    vCardFrame::OnAddCard)
    EVT_MENU(vCard_Edit_Edit,   vCardFrame::OnEditCard)
    EVT_MENU(vCard_Edit_Delete, vCardFrame::OnDeleteCard)

    EVT_MENU(vCard_Help_About,  vCardFrame::OnAbout)

    EVT_UPDATE_UI_RANGE(vCard_Edit_Edit, vCard_Edit_Delete,
                        vCardFrame::OnUpdateEditMenu)

    EVT_LIST_ITEM_ACTIVATED(vCard_ListCtrl, vCardFrame::OnListItemActivated)

    EVT_CLOSE(vCardFrame::OnClose)
END_EVENT_TABLE()

// create a new application object: this macro will allow wxWindows to create
// the application object during program execution (it's better than using a
// static object for many reasons) and also declares the accessor function
// wxGetApp() which will return the reference of the right type (i.e. vCardApp
// and not wxApp)
IMPLEMENT_APP(vCardApp)

// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// vCardApp: the application class
// ----------------------------------------------------------------------------

// main() equivalent: the program execution "starts" here and continues
// normally if OnInit() returns TRUE, returning FALSE aborts the program
// execution
bool vCardApp::OnInit()
{
    // the application name is used in various messages and by default is
    // just the base name of the executable - which is almost ok for us
    // except for the case, so set it manually
    SetAppName(_T("wxCard"));

    // we want to work with all image formats
    wxInitAllImageHandlers();

    // parse the command line: we might have also used a static array
    // containing the entries for all accepted options
    wxCmdLineParser parser(argc, argv);
    parser.AddSwitch(_T("v"), _T("verbose"), _("increase verbosity"));
    parser.AddOption(_T("l"), _T("lang"), _("specify the language to use"));
    parser.AddSwitch(_T("h"), _T("help"), _("give this help message"),
                     wxCMD_LINE_OPTION_HELP);
    parser.AddParam(
                    _("vCard file"),            // parameter description
                    wxCMD_LINE_VAL_STRING,      // type (it is a string)
                    wxCMD_LINE_PARAM_OPTIONAL   // and it can be omitted
                   );

    wxString filename;  // the name of vCard file to open
    switch ( parser.Parse() )
    {
            // cmd line parsed successfully, look what we've got here
        case 0:
            {
                if ( parser.Found(_T("v")) )
                {
                    // make us verbose
                    wxLog::GetActiveTarget()->SetVerbose();
                }

                wxString lang;
                if ( parser.Found(_T("l"), &lang) )
                {
                    // change the language used
                    m_locale = new wxLocale(lang, lang);
                    m_locale->AddCatalog(_T("wxCard"));
                }

                // any parameters?
                if ( parser.GetParamCount() > 0 )
                {
                    // remember the parameter, we'll pass it to the frame later
                    filename = parser.GetParam(0u);
                }
            }
            break;

            // syntax error
        default:
            wxLogError(_("Incorrect syntax, aborting."));
            // fall through

        case -1:
            // help was already given, just terminate by returning FALSE from
            // OnInit()
            return FALSE;
    }

    // create the main application window
    vCardFrame *frame = new vCardFrame;
    if ( filename.length() )
    {
        // show the initially selected card file
        frame->OpenFile(filename);
    }

    frame->Show(TRUE);

    // success: wxApp::OnRun() will be called which will enter the main message
    // loop and the application will run - it will exit automatically when the
    // last top level application window is closed
    return TRUE;
}

int vCardApp::OnExit()
{
    // it might be NULL, but it's ok
    delete m_locale;

    return 0;
}

// ----------------------------------------------------------------------------
// vCardFrame: the main application window
// ----------------------------------------------------------------------------

// ctor: create the frame and its child controls
vCardFrame::vCardFrame()
          : wxFrame(
                    NULL,                   // no parent
                    -1,                     // default id
                    "",                     // title (will be set later)
                    wxDefaultPosition,
                    wxSize(400, 250)
                   )
{
    // create the menu
    wxMenu *menuFile = new wxMenu("", wxMENU_TEAROFF);
    menuFile->Append(vCard_File_New, _("&New\tCtrl-N"), _("Create new card file"));
    menuFile->Append(vCard_File_Open, _("&Open...\tCtrl-O"), _("Open a card file"));
    menuFile->Append(vCard_File_Save, _("&Save\tCtrl-S"), _("Save the card file"));
    menuFile->AppendSeparator();
    menuFile->Append(vCard_File_Quit, _("E&xit\tAlt-X"), _("Quit this program"));

    wxMenu *menuEdit = new wxMenu("", wxMENU_TEAROFF);
    menuEdit->Append(vCard_Edit_Add, _("&Add card..."), _("Add a new card to the file"));
    menuEdit->Append(vCard_Edit_Edit, _("&Edit card..."), _("Edit the selected card"));
    menuEdit->Append(vCard_Edit_Delete, _("&Delete card"), _("Delete the selected card"));

    wxMenu *menuHelp = new wxMenu("", wxMENU_TEAROFF);
    menuHelp->Append(vCard_Help_About, _("&About...\tF1"), _("Show about dialog"));

    wxMenuBar *menuBar = new wxMenuBar;
    menuBar->Append(menuFile, _("&File"));
    menuBar->Append(menuEdit, _("&Edit"));
    menuBar->Append(menuHelp, _("&Help"));

    SetMenuBar(menuBar);

    // create the toolbar
    wxToolBar *tbar = CreateToolBar(wxNO_BORDER | wxTB_FLAT | wxTB_DOCKABLE);
    tbar->AddTool(vCard_File_Open, wxBITMAP(tb_open), _("Open file"));
    tbar->AddSeparator();
    tbar->AddTool(vCard_Edit_Add, wxBITMAP(tb_new), _("Create new card"));
    tbar->AddTool(vCard_Edit_Delete, wxBITMAP(tb_delete), _("Delete card"));
    tbar->AddSeparator();
    tbar->AddTool(vCard_Help_About, wxBITMAP(tb_about), _("About this program"));

    tbar->Realize();    // call this in the end of toolbar construction

    // create the list control taking the entire client area
    m_listctrl = new wxListCtrl(this, vCard_ListCtrl,
                                wxDefaultPosition, wxDefaultSize,
                                wxLC_REPORT | wxRAISED_BORDER);
    m_listctrl->InsertColumn(Column_Name, _("Name"));
    m_listctrl->InsertColumn(Column_Email, _("EMail"));
    m_listctrl->Move(0, 0);

    // create the status bar
    CreateStatusBar();

    // finally set the frame icon an title
    SetIcon(wxICON(wxcard));
    SetFilename("");
}

vCardFrame::~vCardFrame()
{
    WX_CLEAR_ARRAY(m_cards);
}

// operations

void vCardFrame::OpenFile(const wxString& filename)
{
    if ( !wxFile::Exists(filename) )
    {
        wxLogError(_("Can't open file '%s': no such file."));

        return;
    }

    if ( !Save() )
        return;

    // read the file
    m_cards = wxVCard::CreateFromFile(filename);

    // and populate the list ctrl with the cards
    m_listctrl->DeleteAllItems();
    bool empty = TRUE;
    size_t nCards = m_cards.GetCount();
    for ( size_t n = 0; n < nCards; n++ )
    {
        if ( DoAddCard(m_cards[n]) )
            empty = FALSE;
    }

    // adjust the column widths
    size_t nCols = m_listctrl->GetColumnCount();
    for ( size_t nCol = 0; nCol < nCols; nCol++ )
    {
        // if we have any items, let the column adjust itself, otherwise use
        // some default values
        m_listctrl->SetColumnWidth(nCol, empty ? 100 : wxLIST_AUTOSIZE);
    }

    SetFilename(filename);
}

bool vCardFrame::Save()
{
    if ( !m_dirty )
        return TRUE;

    switch ( wxMessageBox(
                          _("Do you want to save changes to the current file?"),
                          _("wxCard: Save file?"),
                          wxYES_NO | wxCANCEL | wxICON_QUESTION
                         ) )
    {
        case wxYES:
            return DoSave();

        case wxNO:
            return TRUE;

        default:
        case wxCANCEL:
            return FALSE;
    }
}

void vCardFrame::SetFilename(const wxString& filename)
{
    m_filename = filename;
    wxString title = wxTheApp->GetAppName();
    if ( m_filename.length() )
    {
        // we have a loaded file
        title << ": " << m_filename;

        SetStatusText(wxString::Format(_("Opened file '%s'"), m_filename.c_str()));
    }
    else
    {
        SetStatusText(_("New file"));
    }

    SetTitle(title);

    // reset the dirty flag in any case as the file changed
    m_dirty = FALSE;
}

long vCardFrame::GetSelection() const
{
    return m_listctrl->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
}

bool vCardFrame::DoSave()
{
    // make sure we have a filename
    if ( !m_filename )
    {
        m_filename = wxSaveFileSelector(
                                        VCARD_NAME,
                                        VCARD_EXT,
                                        NULL,
                                        this
                                       );

        if ( !m_filename )
        {
            // cancelled by user
            return FALSE;
        }
    }

    // do save it
    wxFFile file(m_filename, "w");
    bool ok = file.IsOpened();
    size_t nCards = m_cards.GetCount();
    for ( size_t n = 0; n < nCards && ok; n++ )
    {
        if ( !file.Write(m_cards[n]->Write()) )
        {
            ok = FALSE;
        }
    }

    if ( !ok || !file.Close() )
    {
        wxLogError(_("Failed to save vCard."));

        return FALSE;
    }

    SetStatusText(wxString::Format(_("File '%s' saved"), m_filename.c_str()));

    // no changes any more
    m_dirty = FALSE;

    return TRUE;
}

bool vCardFrame::DoAddCard(const wxVCard *vcard)
{
    wxString name;
    if ( !vcard->GetFullName(&name) )
    {
        wxLogWarning(_("vCard without name ignored."));

        return FALSE;
    }

    long n = m_listctrl->GetItemCount();
    m_listctrl->InsertItem(n, name);

    wxString email;
    if ( vcard->GetEMail(&email) )
    {
        m_listctrl->SetItem(n, Column_Email, email);
    }

    return TRUE;
}

void vCardFrame::DoEdit(long n)
{
    wxCHECK_RET( (size_t)n < m_cards.GetCount(), _T("invalid card index") );

    if ( wxEditVCard(m_cards[n], this) )
    {
        // the card changed
        m_dirty = TRUE;
    }
}

void vCardFrame::DoDelete(long n)
{
    wxCHECK_RET( (size_t)n < m_cards.GetCount(), _T("invalid card index") );

    m_cards.RemoveAt(n);
    m_listctrl->DeleteItem(n);

    // the file changed
    m_dirty = TRUE;
}

// menu callbacks

void vCardFrame::OnNewFile(wxCommandEvent& event)
{
    if ( !Save() )
        return;

    m_listctrl->DeleteAllItems();

    SetFilename("");
}

void vCardFrame::OnSave(wxCommandEvent& event)
{
    (void)DoSave();
}

void vCardFrame::OnOpenFile(wxCommandEvent& event)
{
    // get the file name from user
    wxString filename = wxLoadFileSelector(
                                           VCARD_NAME,
                                           VCARD_EXT,
                                           NULL,
                                           this
                                          );

    if ( filename.length() )
    {
        OpenFile(filename);
    }
    //else: cancelled by user
}

void vCardFrame::OnAbout(wxCommandEvent& event)
{
    (void)wxMessageBox(_("wxCard: vCard editor/viewer written in wxWindows\n(c) 2000 Vadim Zeitlin"),
                       _("About wxCard"),
                       wxICON_INFORMATION);
}

void vCardFrame::OnQuit(wxCommandEvent& event)
{
    // close the frame: this will close the whole program as well
    Close();
}

void vCardFrame::OnAddCard(wxCommandEvent& event)
{
    wxVCard *vcard = wxCreateVCard(this);
    if ( vcard )
    {
        m_cards.Add(vcard);
        DoAddCard(vcard);

        m_dirty = TRUE;
    }
}

void vCardFrame::OnEditCard(wxCommandEvent& event)
{
    DoEdit(GetSelection());
}

void vCardFrame::OnDeleteCard(wxCommandEvent& event)
{
    DoDelete(GetSelection());
}

void vCardFrame::OnListItemActivated(wxListEvent& event)
{
    DoEdit(event.GetIndex());
}

void vCardFrame::OnClose(wxCloseEvent& event)
{
    if ( Save() )
    {
        Destroy();
    }
    else
    {
        // don't close
        event.Veto();
    }
}

void vCardFrame::OnUpdateEditMenu(wxUpdateUIEvent& event)
{
    // the edit and delete operations only make sense if there is a selected
    // card
    event.Enable( GetSelection() != -1 );
}
